@@ -1,6 +1,6 @@
 /* channel-func.c -- SSH channel manipulation functions.
  *
- * Copyright (C) 2013, 2014, 2015 Artyom V. Poptsov <poptsov.artyom@gmail.com>
+ * Copyright (C) 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020 Artyom V. Poptsov <poptsov.artyom@gmail.com>
  *
  * This file is part of Guile-SSH.
  *
@@ -19,7 +19,6 @@
  */
 
 #include <config.h>
-
 #include <assert.h>
 #include <libguile.h>
 #include <libssh/libssh.h>
@@ -51,6 +50,7 @@ Return value is undefined.\
   int res;
   GSSH_VALIDATE_CHANNEL_DATA (data, channel, FUNC_NAME);
   res = ssh_channel_open_session (data->ssh_channel);
+  _gssh_log_debug_format(FUNC_NAME, channel, "result: %d", res);
   if (res != SSH_OK)
     {
       ssh_session session = ssh_channel_get_session (data->ssh_channel);
@@ -80,6 +80,8 @@ Run a shell command CMD without an interactive shell.\
 
   c_cmd = scm_to_locale_string (cmd);
   res = ssh_channel_request_exec (data->ssh_channel, c_cmd);
+  _gssh_log_debug_format(FUNC_NAME, scm_list_2 (channel, cmd),
+                         "result: %d", res);
   free (c_cmd);
   if (res != SSH_OK)
     {
@@ -108,6 +110,7 @@ returned (yet). \
 
   cd = _scm_to_channel_data (channel);
   res = ssh_channel_get_exit_status (cd->ssh_channel);
+  _gssh_log_debug_format(FUNC_NAME, channel, "result: %d", res);
   if (res == SSH_ERROR) {
     _gssh_log_warning (FUNC_NAME,
                        "Could not get exit status",
@@ -136,6 +139,8 @@ Return value is undefined.\
 
   res = ssh_channel_request_send_exit_status (cd->ssh_channel,
                                               scm_to_uint32 (exit_status));
+  _gssh_log_debug_format(FUNC_NAME, scm_list_2 (channel, exit_status),
+                         "result: %d", res);
   if (res != SSH_OK)
     {
       ssh_session session = ssh_channel_get_session (cd->ssh_channel);
@@ -160,6 +165,7 @@ Return value is undefined.\
   GSSH_VALIDATE_OPEN_CHANNEL (channel, SCM_ARG1, FUNC_NAME);
 
   res = ssh_channel_request_pty (data->ssh_channel);
+  _gssh_log_debug_format(FUNC_NAME, channel, "result: %d", res);
   if (res != SSH_OK)
     {
       ssh_session session = ssh_channel_get_session (data->ssh_channel);
@@ -184,6 +190,7 @@ Return value is undefined.\
   GSSH_VALIDATE_OPEN_CHANNEL (channel, SCM_ARG1, FUNC_NAME);
 
   res = ssh_channel_request_shell (data->ssh_channel);
+  _gssh_log_debug_format(FUNC_NAME, channel, "result: %d", res);
   if (res != SSH_OK)
     {
       ssh_session session = ssh_channel_get_session (data->ssh_channel);
@@ -216,6 +223,8 @@ Return value is undefined.\
   c_name  = scm_to_locale_string (name);
   c_value = scm_to_locale_string (value);
   res = ssh_channel_request_env (data->ssh_channel, c_name, c_value);
+  _gssh_log_debug_format(FUNC_NAME, scm_list_3 (channel, name, value),
+                         "result: %d", res);
   if (res != SSH_OK)
     {
       ssh_session session = ssh_channel_get_session (data->ssh_channel);
@@ -231,6 +240,7 @@ Return value is undefined.\
  */
 SCM
 _ssh_result_to_symbol (int res)
+#define FUNC_NAME "_ssh_result_to_symbol"
 {
   switch (res)
     {
@@ -240,12 +250,17 @@ _ssh_result_to_symbol (int res)
       return scm_from_locale_symbol ("again");
     case SSH_ERROR:
       return scm_from_locale_symbol ("error");
+    case SSH_EOF:
+      return scm_from_locale_symbol ("eof");
 
     default:                    /* Must not happen. */
+      _gssh_log_error_format(FUNC_NAME, SCM_BOOL_F,
+                             "Unknown SSH result: %d", res);
       assert (0);
       return SCM_BOOL_F;
     }
 }
+#undef FUNC_NAME
 
 SCM_GSSH_DEFINE (guile_ssh_channel_open_forward,
                  "%channel-open-forward", 5,
@@ -274,7 +289,10 @@ SCM_GSSH_DEFINE (guile_ssh_channel_open_forward,
   res = ssh_channel_open_forward (cd->ssh_channel,
                                   c_remote_host, scm_to_int32 (remote_port),
                                   c_source_host, scm_to_int32 (local_port));
-
+  _gssh_log_debug_format(FUNC_NAME, scm_list_5 (channel, remote_host,
+                                                remote_port, source_host,
+                                                local_port),
+                         "result: %d", res);
   if (res == SSH_OK) {
     SCM_SET_CELL_TYPE (channel, SCM_CELL_TYPE (channel) | SCM_OPN);
   } else {
@@ -317,6 +335,8 @@ SCM_GSSH_DEFINE (guile_ssh_channel_listen_forward,
                             c_address,
                             scm_to_int (port),
                             &bound_port);
+  _gssh_log_debug_format(FUNC_NAME, scm_list_3 (session, address, port),
+                         "result: %d", res);
   if (res != SSH_OK)
     bound_port = -1;
   else if (scm_zero_p (port))
